from __future__ import annotations

from typing import Union


class MRange:
    """A range of values for a parameter in a MATLAB style.
    The range is inclusive of the start and stop values, and the step is the difference between items in the range.
    """
    def __init__(self, start, stop, step):
        self.start = start
        self.stop = stop
        self.step = step
        if self.start == self.stop:
            raise ValueError(f'MRange start and stop values are equal: {self.start} '
                             f'`list(MRange)` will be empty! Use a `Singular({self.start}, 1)` range instead.')
        if self.step == 0:
            raise ZeroDivisionError('MRange step cannot be zero')

    def __eq__(self, other):
        return self.start == other.start and self.stop == other.stop and self.step == other.step

    @property
    def min(self):
        return self.start

    @property
    def max(self):
        return self.stop

    def __iter__(self):
        def range_gen(start, stop, step):
            v = start
            i = 0
            while (step > 0 and v + step <= stop) or (step < 0 and v + step >= stop):
                v = i * step + start
                i += 1
                yield v
        return range_gen(self.start, self.stop, self.step)

    def __getitem__(self, index: int):
        if index < 0 or index >= len(self):
            raise IndexError(f'Index {index} out of range')
        return index * self.step + self.start

    def __str__(self):
        return f'{self.start}:{self.step}:{self.stop}'

    def __repr__(self):
        return f'MStyleRange({self})'

    def __len__(self):
        return int((self.stop - self.start) / self.step) + 1

    @classmethod
    def from_str(cls, string):
        """Parse a string in MATLAB style into a range.
        The string should be of the form start:step:stop
        """
        def float_or_int(s):
            try:
                return int(s)
            except ValueError:
                pass
            return float(s)

        if string.count(':') > 2:
            raise ValueError(f'Range string {string} contains more than two colons')
        step = '1'
        if ':' not in string:
            start, stop = string, string
        elif string.count(':') == 1:
            start, stop = string.split(':')
        else:
            start, step, stop = string.split(':')
        return cls(float_or_int(start), float_or_int(stop), float_or_int(step))


class Singular:
    """A singular range parameter for use with other range parameters in, e.g., a zip.

    Note:
        The Singular range value will be repeated up to `maximum` times in an iterator.
        If `maximum` is None, the Singular range will be repeated forever.
        Therefore, care must be taken to ensure that the Singular range is used in a zip with a range that is
        not infinite.
    """
    def __init__(self, value, maximum=None):
        self.value = value
        self.maximum = maximum

    def __eq__(self, other):
        return self.value == other.value and self.maximum == other.maximum

    def __str__(self):
        return f'{self.value}(up to {self.maximum} times)'

    def __repr__(self):
        return f'Singular({self.value}, {self.maximum})'

    @property
    def min(self):
        return self.value

    @property
    def max(self):
        return self.value

    def __iter__(self):
        def forever():
            while True:
                yield self.value

        def until():
            i = 0
            while i < self.maximum:
                i += 1
                yield self.value

        return until() if self.maximum is not None else forever()

    def __len__(self):
        return self.maximum

    @classmethod
    def from_str(cls, string):
        def float_or_int_or_str(s):
            try:
                return int(s)
            except ValueError:
                pass
            try:
                return float(s)
            except ValueError:
                return s

        if string.count(':') > 0:
            raise ValueError(f'Singular string {string} contains a colon')
        return cls(float_or_int_or_str(string))


def parse_list(range_type, unparsed: list[str]):
    ranges = {}
    while len(unparsed):
        if '=' in unparsed[0]:
            k, v = unparsed[0].split('=', 1)
            ranges[k.lower()] = range_type.from_str(v)
        elif len(unparsed) > 1 and '=' not in unparsed[1]:
            ranges[unparsed[0].lower()] = range_type.from_str(unparsed[1])
            del unparsed[1]
        else:
            raise ValueError(f'Invalid parameter: {unparsed[0]}')
        del unparsed[0]
    return ranges


def parameters_to_scan(parameters: dict[str, Union[list, MRange, Singular]], grid: bool = False):
    """Convert a dictionary of ranged parameters to a list of parameter names and an iterable of parameter value tuples.

    The ranged parameters can be either MRange objects or lists of values. If a list of values is provided, it will be
    iterated over directly.

    :parameter parameters: A dictionary of ranged parameters.
    :parameter grid: Controls how the parameters are iterated; True implies a grid scan, False implies a linear scan.
    """
    if grid:
        for k, v in parameters.items():
            if isinstance(v, Singular):
                parameters[k] = Singular(v.value, 1)

    names = [x.lower() for x in parameters.keys()]
    values = [x if hasattr(x, '__iter__') else [x] for x in parameters.values()]
    if not len(values):
        return 0, names, []
    elif grid:
        from itertools import product
        from math import prod
        # singular MRange objects *should* stop the grid along their axis:
        n_pts = prod([len(v) for v in values])
        return n_pts, names, product(*values)
    else:
        # replace singular MRange entries with Singular iterators, to avoid stopping the zip early:
        n_max = max([len(v) for v in values])
        for i, v in enumerate(values):
            if len(v) > 1 and len(v) != n_max:
                oth = [names[i] for i, n in enumerate(values) if len(n) == n_max]
                par = 'parameters' if len(oth) > 1 else 'parameter'
                have = 'have' if len(oth) > 1 else 'has'
                raise ValueError(f'Parameter {names[i]} has {len(v)} values, but {par} {", ".join(oth)} {have} {n_max}')
        return n_max, names, zip(*[v if len(v) > 1 else Singular(v[0] if isinstance(v, MRange) else v.value, n_max) for v in values])


def _MRange_or_Singular(s: str):
    if ':' in s:
        return MRange.from_str(s)
    return Singular.from_str(s)


def parse_command_line_parameters(unparsed: list[str]) -> dict[str, Union[Singular, MRange]]:
    """Parse a list of input parameters into a dictionary of MRange objects.

    :parameter unparsed: A list of parameters.
    """
    ranges = {}
    index = 0
    while index < len(unparsed):
        if '=' in unparsed[index]:
            k, v = unparsed[index].split('=', 1)
            ranges[k.lower()] = _MRange_or_Singular(v)
        elif index + 1 < len(unparsed) and '=' not in unparsed[index + 1]:
            ranges[unparsed[index].lower()] = _MRange_or_Singular(unparsed[index + 1])
            index += 1
        else:
            raise ValueError(f'Invalid parameter: {unparsed[index]}')
        index += 1
    return ranges


def parse_scan_parameters(unparsed: list[str]) -> dict[str, MRange | Singular]:
    """Parse a list of input parameters into a dictionary of MRange or Singular objects.

    :parameter unparsed: A list of parameters.
    :return: A dictionary of MRange or Singular objects. The Singular objects have their maximum length set to the
             maximum iterations of all the ranges to avoid infinite iterations.
    """
    ranges = parse_command_line_parameters(unparsed)
    max_length = max(len(v) if isinstance(v, MRange) else 1 for v in ranges.values())
    for k, v in ranges.items():
        if isinstance(v, Singular) and v.maximum is None:
            ranges[k] = Singular(v.value, max_length)
    return ranges



import unittest


def parameters(names: tuple[str, ...]):
    from itertools import product
    return tuple(x + y for x, y in product(names, ('speed', 'phase')))


CHOPPER_NAMES = ('pulse_shaping', 'frame_overlap', 'bandwidth')
CHOPPERS = tuple(f'{name}_chopper_{no}' for name in CHOPPER_NAMES for no in (1, 2))
OLD_CHOPPERS = tuple(f'{name}{no}' for name in ('ps', 'fo', 'bw') for no in (1, 2))


class BIFROSTEnergyTestCase(unittest.TestCase):
    def setUp(self):
        from mccode_antlr.loader import parse_mcstas_instr
        instr = f"""DEFINE INSTRUMENT this_IS_NOT_BIFROST(
        pulse_shaping_chopper_1speed, pulse_shaping_chopper_1phase, pulse_shaping_chopper_2speed, pulse_shaping_chopper_2phase, frame_overlap_chopper_1speed, frame_overlap_chopper_1phase, bandwidth_chopper_1speed, bandwidth_chopper_1phase, bandwidth_chopper_2speed, bandwidth_chopper_2phase
        )
        TRACE
        COMPONENT origin = Arm() AT (0, 0, 0) ABSOLUTE
        COMPONENT pulse_shaping_chopper_1 = DiskChopper(theta_0=170, radius=0.35, nu=pulse_shaping_chopper_1speed, phase=pulse_shaping_chopper_1phase) AT (0, 0, 1) RELATIVE PREVIOUS
        COMPONENT pulse_shaping_chopper_2 = DiskChopper(theta_0=170, radius=0.35, nu=pulse_shaping_chopper_2speed, phase=pulse_shaping_chopper_2phase) AT (0, 0, 0.02) RELATIVE pulse_shaping_chopper_1
        COMPONENT frame_overlap_chopper_1 = DiskChopper(theta_0=110, radius=0.35, nu=frame_overlap_chopper_1speed, phase=frame_overlap_chopper_1phase) AT (0, 0, 12) RELATIVE pulse_shaping_chopper_2
        COMPONENT frame_overlap_chopper_2 = DiskChopper(theta_0=115, radius=0.35, nu=frame_overlap_chopper_1speed, phase=frame_overlap_chopper_1phase) AT (0, 0, 4) RELATIVE frame_overlap_chopper_1
        COMPONENT bandwidth_chopper_1 = DiskChopper(theta_0=110, radius=0.35, nu=bandwidth_chopper_1speed, phase=bandwidth_chopper_1phase) AT (0, 0, 80) RELATIVE frame_overlap_chopper_2
        COMPONENT bandwidth_chopper_2 = DiskChopper(theta_0=115, radius=0.35, nu=bandwidth_chopper_2speed, phase=bandwidth_chopper_2phase) AT (0, 0, 0.02) RELATIVE bandwidth_chopper_1
        COMPONENT sample = Arm() AT (0, 0, 80) RELATIVE bandwidth_chopper_2
        END
        """
        self.instr = parse_mcstas_instr(instr)

    def test_names(self):
        from restage.energy import get_energy_parameter_names
        energy_names = get_energy_parameter_names(self.instr.name)
        for name in ('e', 'energy', 'ei', 'wavelength', 'lambda', 'time', 't', 'order'):
            self.assertTrue(name in energy_names)

    def test_parameters_to_scan(self):
        from restage.range import MRange, Singular, parameters_to_scan
        order = Singular(14, 1)
        time = MRange(0.0001, 0.002248, 0.0002)
        ei = MRange(1.7, 24.7, 0.5)
        all_order = list(order)
        all_times = list(time)
        all_ei = list(ei)
        self.assertEqual(len(all_order), 1)
        self.assertEqual(len(all_times), 11)
        self.assertEqual(len(all_ei), 47)
        for x, y in zip(all_order, range(1)):
            o = y * 1 + 14
            self.assertAlmostEqual(x, o)
        for x, y in zip(all_times, range(11)):
            t = y * 0.0002 + 0.0001
            self.assertAlmostEqual(x, t)
        for x, y in zip(all_ei, range(47)):
            e = y * 0.5 + 1.7
            self.assertAlmostEqual(x, e)

        scan_parameters = dict(order=order, time=time, ei=ei)
        npts, names, points = parameters_to_scan(scan_parameters, grid=True)
        self.assertEqual(npts, 47*11)
        self.assertEqual(names, ['order', 'time', 'ei'])
        all_points = list(points)
        self.assertEqual(len(all_points), 47*11)
        for point in points:
            self.assertEqual(len(point), 3)

        # the orientation of the grid is not super important, but it should be consistent
        # with the order of the parameters
        for i, point in enumerate(points):
            row = i // len(all_times)
            col = i % len(all_times)
            self.assertEqual(point[0], all_order[0])
            self.assertAlmostEqual(point[1], all_times[col])
            self.assertAlmostEqual(point[2], all_ei[row])

    def test_translator(self):
        from restage.energy import energy_to_chopper_translator
        from restage.energy import bifrost_translate_energy_to_chopper_parameters
        from restage.range import MRange, Singular, parameters_to_scan


        translator = energy_to_chopper_translator(self.instr.name)
        self.assertEqual(translator, bifrost_translate_energy_to_chopper_parameters)

        order = Singular(14,  1)
        time = MRange(0.0001, 0.002248, 0.0002)
        ei = MRange(1.7, 24.7, 0.5)
        scan_parameters = dict(order=order, time=time, ei=ei)

        spts, names, points = parameters_to_scan(scan_parameters, grid=True)

        self.assertEqual(47*11, spts)
        self.assertEqual(names, ['order', 'time', 'ei'])

        chopper_parameters = parameters(CHOPPERS)
        for point in points:
            kv = {k: v for k, v in zip(names, point)}
            translated = translator(kv)
            for x in chopper_parameters:
                self.assertTrue(x in translated)

            self.assertEqual(len(translated), len(chopper_parameters))
            self.assertAlmostEqual(translated['bandwidth_chopper_1speed'], 14.0)
            self.assertAlmostEqual(translated['bandwidth_chopper_2speed'], -14.0)
            self.assertAlmostEqual(translated['frame_overlap_chopper_1speed'], 14.0)
            self.assertAlmostEqual(translated['frame_overlap_chopper_1speed'], 14.0)
            self.assertAlmostEqual(translated['pulse_shaping_chopper_1speed'], 14*14.0)
            self.assertAlmostEqual(translated['pulse_shaping_chopper_2speed'], 14*14.0)

    def test_calculations(self):
        from itertools import product
        from chopcal import bifrost as mcstas_bifrost_calculation
        from restage.energy import bifrost_translate_energy_to_chopper_parameters

        shortest_time = 0.0001  # this is approximately twice the opening time of the pulse shaping choppers at 15*14 Hz
        # Normal operation  Shortest full-height pulse  Shorter pulses reduce height
        #      /-----\                  /\
        # ----/       \---  -----------/  \------------ -------------/\--------------

        order = 14  # the McStas calculations are for 14th order *only* -- though they can be reduced to lower orders

        # the longest time has both disks (nearly) in phase [in phase if no distance between them]
        # but we reduce that here to ensure the McStas calculation does not reduce the order
        longest_time = (170 / 360) / order / 14 - shortest_time

        smallest_energy = 0.75  # ~4 full source periods to reach the sample, and more than 1 meV energy gain
        largest_energy = 25.  # a guess, but depends on the source spectra

        n_time, n_energy = 100, 100
        d_time, d_energy = (longest_time - shortest_time) / n_time, (largest_energy - smallest_energy) / n_energy
        for time_index, energy_index in product(range(n_time), range(n_energy)):
            time = shortest_time + time_index * d_time
            energy = smallest_energy + energy_index * d_energy

            kv = {'order': order, 'time': time, 'ei': energy}
            translated = bifrost_translate_energy_to_chopper_parameters(kv)
            from_mcstas = mcstas_bifrost_calculation(energy, 0., time)
            for o, x in zip(OLD_CHOPPERS, CHOPPERS):
                # chopcal >= 0.4.0 returns a dictionary of Chopper objects
                chopper = from_mcstas[o]
                for prop in ('speed', 'phase'):
                    self.assertAlmostEqual(getattr(chopper, prop), translated[f'{x}{prop}'])


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import unittest
from mccode_antlr.compiler.check import simple_instr_compiles


def compiled_test(method, compiler: str | None = None):
    if compiler is None:
        compiler = 'cc'
    if simple_instr_compiles(compiler):
        return method

    @unittest.skip(f"Skipping due to lack of working {compiler}")
    def skipped_method(*args, **kwargs):
        return method(*args, **kwargs)

    return skipped_method


def gpu_compiled_test(method):
    return compiled_test(method, 'acc')


def mpi_compiled_test(method):
    return compiled_test(method, 'mpi/cc')


class SingleTestCase(unittest.TestCase):
    def setUp(self):
        from restage.splitrun import make_splitrun_parser
        self.parser = make_splitrun_parser()

    def test_parsing(self):
        args = self.parser.parse_args(['test.instr', 'a=1', 'b=2', '--split-at=here', '-m'])
        self.assertEqual(args.instrument, 'test.instr')
        self.assertEqual(args.parameters, ['a=1', 'b=2'])
        self.assertEqual(args.split_at, 'here')
        self.assertTrue(args.mesh)

    def test_mixed_parsing(self):
        from mccode_antlr.run.runner import sort_args
        args = self.parser.parse_args(sort_args(['test.instr', '-m', 'a=1', 'b=2', '--split-at=here']))
        self.assertEqual(args.instrument, 'test.instr')
        self.assertEqual(args.parameters, ['a=1', 'b=2'])
        self.assertEqual(args.split_at, 'here')
        self.assertTrue(args.mesh)

    def test_negative_count_throws(self):
        from restage.splitrun import si_int, si_int_limits
        with self.assertRaises(ValueError):
            si_int('-10')
        with self.assertRaises(ValueError):
            si_int_limits('10--4+10') # '10--4+10' -> ('10', '-4+10') -> (10, '-4', '10')
        with self.assertRaises(ValueError):
            # This is probably a parsing error, but would raise a negative error
            # in si_int, so is fine. '-10]-2[-1' -> ('','10]-2[-1') -> Error on int('')
            si_int_limits('-10]-2[-1')

    def test_mccode_flags(self):
        args = self.parser.parse_args(['test.instr', '-s', '123456', '-n', '1', '-d', '/a/dir', '-t', '-g'])
        self.assertEqual(args.seed, 123456)
        self.assertEqual(args.ncount, (None, 1, None))
        self.assertEqual(args.dir, '/a/dir')
        self.assertEqual(args.trace, True)
        self.assertEqual(args.gravitation, True)

        args = self.parser.parse_args(['test.instr', '-s=99999', '-n=10000', '-d=/b/dir'])
        self.assertEqual(args.seed, 99999)
        self.assertEqual(args.ncount, (None, 10000, None))
        self.assertEqual(args.dir, '/b/dir')
        self.assertEqual(args.trace, False)
        self.assertEqual(args.gravitation, False)

        args = self.parser.parse_args(['test.instr', '--seed', '888', '--ncount', '4', '--dir', '/c/dir', '--trace',
                                       '--gravitation', '--bufsiz', '1000', '--format', 'NEXUS'])
        self.assertEqual(args.seed, 888)
        self.assertEqual(args.ncount, (None, 4, None))
        self.assertEqual(args.dir, '/c/dir')
        self.assertEqual(args.trace, True)
        self.assertEqual(args.gravitation, True)
        self.assertEqual(args.bufsiz, 1000)
        self.assertEqual(args.format, 'NEXUS')

        args = self.parser.parse_args(['test.instr', '--seed=777', '--ncount=5', '--dir=/d/dir', '--bufsiz=2000',
                                       '--format=RAW'])
        self.assertEqual(args.seed, 777)
        self.assertEqual(args.ncount, (None, 5, None))
        self.assertEqual(args.dir, '/d/dir')
        self.assertEqual(args.trace, False)
        self.assertEqual(args.gravitation, False)
        self.assertEqual(args.bufsiz, 2000)
        self.assertEqual(args.format, 'RAW')

    def test_ncount_varieties(self):
        args = self.parser.parse_args(['test.instr', '--ncount=5'])
        self.assertEqual(args.ncount, (None, 5, None))
        args = self.parser.parse_args(['test.instr', '-n' ,'4k'])
        self.assertEqual(args.ncount, (None, 4000, None))
        args = self.parser.parse_args(['test.instr', '-n', '3-2+1'])
        self.assertEqual(args.ncount, (3, 2, 1))
        args = self.parser.parse_args(['test.instr', '-n', '1M]1G[1T'])
        self.assertEqual(args.ncount, (10**6, 10**9, 10**12))
        args = self.parser.parse_args(['test.instr', '-n', '1Ki}Mi{2Gi'])
        self.assertEqual(args.ncount, (2**10, 2**20, 2**31))
        args = self.parser.parse_args(['t.instr', '-n', '1.1M', '--nmin', 'M', '--nmax', '2M'])
        self.assertEqual(args.ncount, (None, 1100000, None))
        self.assertEqual(args.nmin, 1000000)
        self.assertEqual(args.nmax, 2000000)


    def test_parameters(self):
        from restage.range import MRange, Singular, parameters_to_scan, parse_scan_parameters
        args = self.parser.parse_args(['test.instr', 'a=1.0', 'b=2', 'c=3:5', 'd=blah', 'e=/data', '-m'])
        self.assertEqual(args.parameters, ['a=1.0', 'b=2', 'c=3:5', 'd=blah', 'e=/data'])
        parameters = parse_scan_parameters(args.parameters)
        self.assertTrue(isinstance(parameters['a'], Singular))
        self.assertTrue(isinstance(parameters['b'], Singular))
        self.assertTrue(isinstance(parameters['c'], MRange))
        self.assertTrue(isinstance(parameters['d'], Singular))
        self.assertTrue(isinstance(parameters['e'], Singular))
        # Singular parameters should have their maximum repetitions set to the longest MRange
        for v in parameters.values():
            self.assertEqual(len(v), 3)
        n_pts, names, scan = parameters_to_scan(parameters)
        self.assertEqual(n_pts, 3)
        self.assertEqual(names, ['a', 'b', 'c', 'd', 'e'])
        for i, values in enumerate(scan):
            self.assertEqual(len(values), 5)
            self.assertEqual(values[0], 1.0)
            self.assertEqual(values[1], 2)
            self.assertEqual(values[2], 3 + i)
            self.assertEqual(values[3], 'blah')
            self.assertEqual(values[4], '/data')

    def test_mcpl_split_parameters(self):
        args = self.parser.parse_args(['test.instr', 'a=1.0', 'b=2', 'c=3:5', 'd=blah',  'e=/data',
                                       '--mcpl-input-parameters', 'preload:1,v_smear:0.01'])
        self.assertEqual(args.parameters, ['a=1.0', 'b=2', 'c=3:5', 'd=blah', 'e=/data'])
        self.assertEqual(args.mcpl_input_parameters, [('preload', '1'), ('v_smear', '0.01')])
        args = self.parser.parse_args(['new_tst.instr', '--mcpl-output-parameters', 'preload:1',
                                       '--mcpl-input-component=MCPL_input_once'])
        self.assertEqual(args.parameters, [])
        self.assertEqual(args.mcpl_output_parameters, [('preload', '1')])
        self.assertEqual(args.mcpl_input_component, 'MCPL_input_once')


class DictWranglingTestCase(unittest.TestCase):
    def test_regularization(self):
        from restage.splitrun import regular_mccode_runtime_dict
        short_names = dict(s=1, n=2, d=3, t=4, g=5, bufsiz=6, format=7)
        long_names = dict(seed=1, ncount=2, dir=3, trace=4, gravitation=5, bufsiz=6, format=7)
        self.assertEqual(regular_mccode_runtime_dict(short_names), long_names)


class SplitRunTestCase(unittest.TestCase):
    def _skip_without_compiler(self):
        import subprocess
        from mccode_antlr.config import config
        try:
            subprocess.run([config['cc'].get(str), '--version'], check=True)
        except FileNotFoundError:
            self.skipTest(f'Compiler {config["cc"]} not found')

    def _skip_without_mcpl(self):
        import subprocess
        try:
            subprocess.run(['mcpl-config', '--version'], check=True)
        except FileNotFoundError:
            self.skipTest('mcpl-config not found')
        except RuntimeError:
            self.skipTest('mcpl-config failed')

    def _skip_checks(self):
        self._skip_without_compiler()
        self._skip_without_mcpl()

    def _define_instr(self):
        from math import pi, asin, sqrt
        from mccode_antlr.loader.loader import parse_mcstas_instr

        d_spacing = 3.355  # (002) for Highly-ordered Pyrolytic Graphite
        mean_energy = 5.0
        energy_width = 1.0
        hbar_sq_over_m = 2.0722  # meV Angstrom^2
        mean_ki = sqrt(mean_energy / hbar_sq_over_m)
        min_ki = sqrt((mean_energy - energy_width) / hbar_sq_over_m)
        max_ki = sqrt((mean_energy + energy_width) / hbar_sq_over_m)
        min_a1 = asin(pi / d_spacing / max_ki) * 180 / pi
        max_a1 = asin(pi / d_spacing / min_ki) * 180 / pi
        instr = f"""
        DEFINE INSTRUMENT splitRunTest(a1=0, a2=0, virtual_source_x=0.05, virtual_source_y=0.1)
        TRACE
        COMPONENT origin = Arm() AT (0, 0, 0) ABSOLUTE
        COMPONENT source = Source_simple(yheight=0.25, xwidth=0.2, dist=1.5, focus_xw=0.06, focus_yh=0.12,
                                         E0={mean_energy}, dE={energy_width})
                           AT (0, 0, 0) RELATIVE origin
        COMPONENT m0 = PSD_monitor(xwidth=0.1, yheight=0.15, nx=100, ny=160, restore_neutron=1) AT (0, 0, 0.01) RELATIVE PREVIOUS
        COMPONENT guide1 = Guide_gravity(w1 = 0.06, h1 = 0.12, w2 = 0.05, h2 = 0.1, l = 15, m = 8) 
                          AT (0, 0, 1.5) RELATIVE  PREVIOUS
        COMPONENT guide1_end = Arm() AT (0, 0, 15) RELATIVE PREVIOUS
        COMPONENT m1 = PSD_monitor(xwidth=0.1, yheight=0.15, nx=100, ny=160, restore_neutron=1) AT (0, 0, 0.01) RELATIVE PREVIOUS
        COMPONENT monitor = E_monitor(xwidth=0.05, yheight=0.1, nE=50,
                                      Emin={mean_energy - 2 * energy_width}, Emax={mean_energy + 2 * energy_width})
                          AT (0, 0, 0.01) RELATIVE PREVIOUS
        COMPONENT image = PSD_monitor(xwidth=0.1, yheight=0.15, nx=100, ny=160) AT (0, 0, 0.01) RELATIVE PREVIOUS
        COMPONENT guide2 = Guide_gravity(w1 = 0.05, h1 = 0.1, l = 15, m = 8) AT (0, 0, 0.01) RELATIVE PREVIOUS
        COMPONENT guide2_end = Arm() AT (0, 0, 15) RELATIVE PREVIOUS
        COMPONENT aperture = Slit(xwidth=virtual_source_x, yheight=virtual_source_y) AT (0, 0, 0.01) RELATIVE PREVIOUS
        COMPONENT before_split = PSD_monitor(xwidth=0.1, yheight=0.15, nx=100, ny=160) AT (0, 0, 0.01) RELATIVE PREVIOUS
        COMPONENT split_at = Arm() AT (0, 0, 0.0001) RELATIVE PREVIOUS
        COMPONENT after_split = PSD_monitor(xwidth=0.1, yheight=0.15, nx=100, ny=160) AT (0, 0, 0.01) RELATIVE PREVIOUS
        COMPONENT mono_point = Arm() AT (0, 0, 0.8) RELATIVE split_at
        COMPONENT mono = Monochromator_curved(zwidth = 0.02, yheight = 0.02, NH = 13, NV = 7, DM={d_spacing}) 
                         AT (0, 0, 0) RELATIVE  mono_point ROTATED (0, a1, 0) RELATIVE mono_point
        COMPONENT sample_arm = Arm() AT (0, 0, 0) RELATIVE mono_point ROTATED (0, a2, 0) RELATIVE mono_point
        COMPONENT detector = Monitor(xwidth=0.01, yheight=0.05) AT (0, 0, 0.8) RELATIVE sample_arm
        END
        """
        return parse_mcstas_instr(instr), min_a1, max_a1

    def setUp(self) -> None:
        from pathlib import Path
        from tempfile import mkdtemp
        self._skip_checks()
        self.instr, self.min_a1, self.max_a1 = self._define_instr()
        with Path().joinpath('splitRunTest.instr').open('w') as file:
            self.instr.to_file(file)
        self.dir = Path(mkdtemp())

    def tearDown(self) -> None:
        import shutil
        if self.dir.exists():
            shutil.rmtree(self.dir)

    def test_simple_scan(self):
        """This test requires MCPL shared libraries to work

        For some unexplored reason, MCPL shared libraries are not found in their
        default installed location, /usr/local/lib64/libmcpl.so; so this test must
        be invoked with that location specified, e.g.,
            $ LD_LIBRARY_PATH=/usr/local/lib64 pytest test/test_single.py -k test_simple_scan
        """
        # Scanning a1 and a2 with a2=2*a1 should produce approximately the same intensity for all points
        # as long as a1 is between the limits of min_a1 and max_a1
        from restage.splitrun import splitrun
        from restage.range import parse_scan_parameters
        # since the source emits a narrow energy bandwidth, we only detect neutrons over a small (a1,a2) range
        scan = parse_scan_parameters([f'a1={self.min_a1}:0.5:{self.max_a1}', f'a2={2*self.min_a1}:{2*self.max_a1}'])

        # The way that McCode handles directories is extremely finicky. If the _actual_ simulation directory
        # exists, the simulation will fail (even if it is empty!), but if the _parent_ directory does not exist,
        # the simulation will fail. So we need to create the parent directory, but not the simulation directory.
        # The real trick is that the simulation directory is a subdirectory of the one specified here.
        output = self.dir.joinpath('test_simple_scan')
        if not output.exists():
            output.mkdir(parents=True)

        # run the scan
        splitrun(self.instr, scan, precision={}, split_at='split_at', grid=False,
                 ncount=10_000,
                 dir=output,
                 mcpl_output_parameters={'weight_mode': '1'},
                 mcpl_input_component='MCPL_input_once',
                 mcpl_input_parameters={'preload': '1'}
                 )

        # check the scan directory for output
        for x in self.dir.glob('**/*.dat'):
            print(x)

        # It would be nice to check that the produced mccode.sim and mccode.dat files look right.

    @mpi_compiled_test
    def test_parallel_scan(self):
        """This test requires mpicc and MCPL shared libraries to work

        On some systems (Fedora at least) specifying which MPI to use requires using
        the `module` system, e.g.,
            $ module load mpi/openmpi-x86_64
        And for some unexplored reason, MCPL shared libraries are not found in their
        default installed location, /usr/local/lib64/libmcpl.so; so this test must
        be invoked with that location specified, e.g.,
            $ LD_LIBRARY_PATH=/usr/local/lib64 pytest test/test_single.py -k test_parallel_scan

        Another potential issue to address, the `MCPL_*.comp` components may use a
        special `@MCPLFLAGS@` directive to read flags from a configuration file instead
        of using the provided `mcpl-config --show buildflags` command.
        You might need to use `mccode-antlr config save -v` to create/locate the
        mccode-antlr `config.yml` file then edit/add the entry for `flags.mcpl` with
        the output of the tool above.
        """
        from restage.splitrun import splitrun
        from restage.range import parse_scan_parameters
        scan = parse_scan_parameters([f'a1={self.min_a1}:0.5:{self.max_a1}', f'a2={2 * self.min_a1}:{2 * self.max_a1}'])
        output = self.dir.joinpath('test_parallel_scan')
        if not output.exists():
            output.mkdir(parents=True)
        splitrun(self.instr, scan, precision={}, split_at='split_at', grid=False, ncount=100_000, dir=output,
                 parallel=True, process_count=4,
                 mcpl_output_parameters={'weight_mode': '1'},
                 mcpl_input_component='MCPL_input_once',
                 mcpl_input_parameters={'preload': '1'}
                 )

        # check the scan directory for output
        for x in self.dir.glob('**/*.dat'):
            print(x)


if __name__ == '__main__':
    unittest.main()



from pathlib import Path


def mcpl_real_filename(filename: Path) -> Path:
    """MCPL_output from McCode instruments has the bad habit of changing the output file name silently.
    Find the _real_ output file name by looking for the expected variants"""
    base, ext = filename.parent / filename.stem, filename.suffix
    if ext in ('.gz',):
        ext = base.suffix + ext
        base = base.parent / base.stem
    extensions = {'.mcpl.gz', '.mcpl', ''}
    if ext not in extensions:
        ValueError(f'Unsupported file extension: {ext}')
    for ext in extensions:
        check = base.with_suffix(ext)
        if check.exists() and check.is_file():
            return check
    raise FileNotFoundError(f'Could not find MCPL file {filename}')


def mcpl_real_extension(filename: Path) -> str:
    for ext in ('.mcpl.gz', '.mcpl'):
        if str(filename).endswith(ext):
            return ext
    return ''


def mcpl_particle_count(filename):
    """Call the MCPL command line tool to get the number of particles in an MCPL file"""
    # There _is_ a Python module for reading MCPL files, but it doesn't close file handles!
    from subprocess import run, PIPE
    import re
    command = ['mcpltool', '--justhead', str(mcpl_real_filename(filename))]
    result = run(command, stdout=PIPE)
    if result.returncode != 0:
        raise RuntimeError(f'mcpltool failed with return code {result.returncode}')
    info = result.stdout.decode('utf-8')
    r = re.compile(r'No\. of particles\s*:\s*(\d+)')
    m = r.search(info)
    if m is None:
        raise RuntimeError(f'Could not find number of particles in {info}')
    return int(m.group(1))


def mcpl_merge_files(files: list[Path], filepath: Path, keep_originals: bool = False):
    """Merge a list of MCPL files into a single file using mcpltool.

    :param files: The list of files to merge.
    :param filename: The name of the output file.
    :param keep_originals: Whether to keep the original files.

    :raises RuntimeError: If mcpltool fails.

    :note: This function is not thread-safe.
        A future version of the Python mcpl package might include a merge function.
    """
    from subprocess import run
    real_filenames = [mcpl_real_filename(f) for f in files]
    # if the real filenames have .mcpl or .mcpl.gz, the merged filename should too
    ext = mcpl_real_extension(real_filenames[0])
    filename = filepath.with_suffix(ext).as_posix()

    command = ['mcpltool', '--merge', filename] + [str(f) for f in real_filenames]
    result = run(command)
    if result.returncode != 0:
        raise RuntimeError(f'mcpltool failed with return code {result.returncode} for command {" ".join(command)}')
    if not keep_originals:
        for file in real_filenames:
            file.unlink()


def mcpl_rename_file(source: Path, dest: Path, strict: bool = False):
    filepath = mcpl_real_filename(source) # this could be '{name}', '{name}.mcpl', or '{name}.mcpl.gz'
    ext = mcpl_real_extension(filepath)

    if not dest.name.endswith(ext):
        if strict:
            raise RuntimeError(f"Destination {dest} does not have extension matching {source}")
        dest = dest.with_suffix(ext)

    filepath.rename(dest)
    return dest
